// Edit from veronica
// Resident Evil/Biohazard: Code: Veronica Auto Splitter
// By Kapdap 2020/07/16
// https://github.com/kapdap/re-cvx-autosplitter

state("pcsx2") {}

startup
{
    // Split tracker
    vars.splits = new Dictionary<string, bool>();

    // Inventory split tracker
    vars.itemSplits = new Dictionary<byte, bool>();
    vars.itemIndex = new Dictionary<byte, string>();

    // Adds a split
    vars.AddSplit = (Action<string, bool, string, string>)((id, default_value, description, parent) => {
        vars.splits.Add(id, false);
        settings.Add(id, default_value, description, parent);
    });

    // Adds an inventory split
    vars.AddItemSplit = (Action<byte, string, bool, string, string>)((code, id, default_value, description, parent) => {
        vars.itemSplits.Add(code, false);
        vars.itemIndex.Add(code, id);  // Map item code to split name
        settings.Add(id, default_value, description, parent);
    });

    // Resets split tracker
    vars.ResetSplits = (Action) (() => {
        List<string> keys = new List<string>(vars.splits.Keys);
        foreach(string key in keys)
            vars.splits[key] = false;
    });

    // Resets inventory split tracker
    vars.ResetItemSplits = (Action) (() => {
        List<byte> keys = new List<byte>(vars.itemSplits.Keys);
        foreach(byte key in keys)
            vars.itemSplits[key] = false;
    });

    // Swaps 2 byte values
    vars.SwapBytes = (Func<ushort, ushort>)((value) => {
        int b1 = (value >> 0) & 0xff;
        int b2 = (value >> 8) & 0xff;

        return (ushort) (b1 << 8 | b2 << 0);
    });

    // Swaps 4 byte values
    vars.SwapBytesInt = (Func<uint, uint>)((value) => {
        return ((value & 0x000000ff) << 24) +
            ((value & 0x0000ff00) << 8) +
            ((value & 0x00ff0000) >> 8) +
            ((value & 0xff000000) >> 24);
    });

    settings.Add("events", true, "Events");
    vars.AddSplit("usesyringe1", true, "Used Syringe 1", "events");
    vars.AddSplit("usesyringe2", true, "Used Syringe 2", "events");
    vars.AddSplit("usesyringe3", true, "Used Syringe 3", "events");

    vars.AddSplit("boss", true, "Boss", "events");
    settings.SetToolTip("boss", "Splits when killed Boss");

    vars.AddSplit("endGame", true, "End Game", "events");
    settings.SetToolTip("endGame", "Splits when the game rank screen displays");

    settings.Add("weapongroup", true, "Weapons");
    vars.AddItemSplit(0x65, "axe", true, "Axe", "weapongroup");
	
    settings.Add("keygroup", true, "Keys");
    vars.AddItemSplit(0x34, "syringe1", true, "Syringe (Solvent) 1", "keygroup");
    vars.AddItemSplit(0x44, "syringe2", true, "Syringe (Solvent) 2", "keygroup");
    vars.AddItemSplit(0x38, "adminkey", true, "Administrator Key", "keygroup");
    vars.AddItemSplit(0x43, "buildingkey", true, "Auxiliary Building Key", "keygroup");

}

init
{
    vars.lastSplit = String.Empty; // Name of the most recent split
    vars.gameProcess = String.Empty; // Used to detect when the process has changed
    vars.productCode = String.Empty; // Used to detect when the game release has changed
    vars.basePointer = IntPtr.Zero; // Emulator virtual memory base pointer
    vars.isBigEndian = false; // Console uses big endian (e.g. PS3 and GCN)

    // Log splits
    vars.LogSplit = (Action<string>)((text) => {
        vars.lastSplit = text;

        string log = timer.CurrentTime[timer.CurrentTimingMethod] + ": " + text;

        if (settings["logSplits"])
            File.AppendAllText(vars.logPath, log + System.Environment.NewLine);

        if (settings["debugSplits"])
            print("recvx split: " + log);
    });

    // Update split status
    vars.UpdateSplit = (Func<string, bool>)((split) => {
        if (!vars.splits[split])
        {
            vars.LogSplit(split);
            vars.splits[split] = true;
            return settings[split];
        }
        else
            return false;
    });

    // Update item split status
    vars.UpdateItemSplit = (Func<byte, bool>)((id) => {
        if (!vars.itemSplits[id])
        {
            vars.LogSplit(vars.itemIndex[id] + " (" + id.ToString() + ")");
            vars.itemSplits[id] = true;
            return settings[vars.itemIndex[id]];
        }
        else
            return false;
    });

    // Sets memory pointers for the detected game release
    vars.UpdatePointers = (Action) (() => {
        switch ((string)vars.productCode)
        {
            default: // SLPM_656.92 [PS2] [JP] Biohazard: outbreak f2
                vars.timePtr = 0x004912B8;
                vars.roomPtr = 0x0047BF12;
                vars.screenPtr = 0x004912A0;
                vars.healthPtr = 0x0047C270;
                vars.statusPtr = 0x0047C8E8;
                vars.pausePtr = 0x004912BE;
                vars.characterPtr = 0x0047C8E0;
                vars.inventoryPtr = 0x0047C9C4;
                vars.bosshpPtr = 0x007612AC;
                break;
        }
    });
	
    // Detects if the process has changed
    vars.UpdateProcess = (Action) (() => {
        // Update process name if application has changed
        if (vars.gameProcess != game.ProcessName)
        {
            vars.gameProcess = game.ProcessName;

            // Set emulator edianess
            switch ((string)vars.gameProcess.ToLower())
            {
                default: //pcsx2
                    vars.isBigEndian = false;
                    break;
            }
        }
    });

    // Set emulator base pointer
    vars.UpdatePointer = (Action) (() => {
        switch ((string)vars.gameProcess.ToLower())
        {
            default: // pcsx2
                vars.basePointer = 0x20000000;
                break;
        }
    });

    // Detects if the game release has changed
    vars.UpdateProduct = (Action) (() => {
        string productCode;

        // Read game product code from memory
        switch ((string)vars.gameProcess.ToLower())
        {
            default: //pcsx2
                productCode = memory.ReadString(new IntPtr(vars.basePointer + 0x00015B90), 11);
                break;
        }

        // Update product code if the game release has changed
        if (vars.productCode != productCode)
        {
            vars.productCode = productCode;
            vars.UpdatePointers();
        }
    });

    // Updates game values
    vars.UpdateValues = (Action) (() => {
        uint time = 0; // In Game Timer
        byte room = 0; // Room the character is in
        byte screen = 0; // Current game screen
        uint health = 0; // Character health
        byte status = 0; // Character status info
        byte pause = 0; // pause status info
        byte character = 0; // Character ID
        ushort bosshp = 0; // bosshp

        // Read values from memory
        memory.ReadValue<uint>(new IntPtr(vars.basePointer + vars.timePtr), out time);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.roomPtr), out room);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.screenPtr), out screen);
        memory.ReadValue<uint>(new IntPtr(vars.basePointer + vars.healthPtr), out health);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.statusPtr), out status);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.pausePtr), out pause);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.characterPtr), out character);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.bosshpPtr), out bosshp);

        current.slot = 0; // Inventory slot number of the equipped item
        current.ammo = 0; // Ammo count for the equipped weapon
        current.time = vars.isBigEndian ? (int)vars.SwapBytesInt(time) : (int)time;
        current.room = room;
        current.screen = screen;
        current.health = vars.isBigEndian ? (int)vars.SwapBytesInt(health) : (int)health;
        current.status = "Normal";
        current.character = character;
        current.inventory = new byte[8]; // Current characters inventory
		current.bosshp = bosshp;

        // Check status for gassed or poison flags
        if ((status & 0x20) != 0)
            current.status = "Gassed";
        else if ((status & 0x08) != 0)
            current.status = "Poison";

        int index = -1; // Inventory array index

        // Pointer to the current characters inventory table
        IntPtr pointer = new IntPtr(vars.basePointer + vars.inventoryPtr-2);

        // Read inventory of the current character
        for (int i = 0; i < 9; ++i)
        {
            int item = 0; // Inventory entry

            // Read inventory entry from memory
            memory.ReadValue<int>(pointer, out item);

            // Move pointer to the next entry
            pointer = IntPtr.Add(pointer, 0x1);

            // First entry is the slot number of the equipped item
            if (i == 0)
            {
                current.slot = item;
                continue;
            }

            // Remaing entries are inventory items
            byte[] bytes = BitConverter.GetBytes(item);

            // Set inventory index to the item ID
            current.inventory[++index] = bytes[1];

            // Read quantity if this item is equipped
            if (current.slot == (index + 1))
                current.ammo = BitConverter.ToInt16(bytes, 0);
        }
    });

    // Initialise values
    vars.UpdateProcess();
    vars.UpdatePointer();
    vars.UpdateProduct();
    vars.UpdateValues();
}

start
{
    // Start when IGT is 0, in the Prison Cell room and the 1st item slot contains the Lighter
    return (current.time > 0);
}

update
{
    vars.UpdateProcess();
    vars.UpdatePointer();
    vars.UpdateProduct();
    vars.UpdateValues();

    if (timer.CurrentPhase == TimerPhase.NotRunning)
    {
        vars.ResetSplits();
        vars.ResetItemSplits();
    }
}

split
{
    // Split when an inventory entry changes
    for (int i = 0; i < current.inventory.Length; i++)
    {
        byte item = current.inventory[i];

        // Continue if the entry has not changed
        if (item == old.inventory[i]) continue;

        switch (item)
        {
            default: // Split the first time an item is added to the inventory
                if (vars.itemSplits.ContainsKey(item))
                    return vars.UpdateItemSplit(item);
                break;
        }
    }

    if (current.room == 15 && current.screen == 0x04)
        return vars.UpdateSplit("usesyringe1");

    if (current.room == 20 && current.screen == 0x04)
        return vars.UpdateSplit("usesyringe2");

    if (current.room == 27 && current.screen == 0x04)
        return vars.UpdateSplit("usesyringe3");

    if (current.room == 16 && current.screen == 0x04) // when boss ded
        return vars.UpdateSplit("boss");

    if (current.screen >= 0x0C) // End Game Rank Screen
        return vars.UpdateSplit("endGame");
}

reset
{
    if (current.room == 0 && current.time == 0)
    {
        return true;
    }
}

gameTime
{
    // Seconds = IGT / 60
    return TimeSpan.FromSeconds(current.time / 0.03 / 1000);
}

isLoading
{
    return true;
}