// Edit from veronica
// Resident Evil/Biohazard: Code: Veronica Auto Splitter
// By Kapdap 2020/07/16
// https://github.com/kapdap/re-cvx-autosplitter

state("pcsx2") {}

startup
{
    // Split tracker
    vars.splits = new Dictionary<string, bool>();

    // Inventory split tracker
    vars.itemSplits = new Dictionary<byte, bool>();
    vars.itemIndex = new Dictionary<byte, string>();

    // Adds a split
    vars.AddSplit = (Action<string, bool, string, string>)((id, default_value, description, parent) => {
        vars.splits.Add(id, false);
        settings.Add(id, default_value, description, parent);
    });

    // Adds an inventory split
    vars.AddItemSplit = (Action<byte, string, bool, string, string>)((code, id, default_value, description, parent) => {
        vars.itemSplits.Add(code, false);
        vars.itemIndex.Add(code, id);  // Map item code to split name
        settings.Add(id, default_value, description, parent);
    });

    // Resets split tracker
    vars.ResetSplits = (Action) (() => {
        List<string> keys = new List<string>(vars.splits.Keys);
        foreach(string key in keys)
            vars.splits[key] = false;
    });

    // Resets inventory split tracker
    vars.ResetItemSplits = (Action) (() => {
        List<byte> keys = new List<byte>(vars.itemSplits.Keys);
        foreach(byte key in keys)
            vars.itemSplits[key] = false;
    });

    // Swaps 2 byte values
    vars.SwapBytes = (Func<ushort, ushort>)((value) => {
        int b1 = (value >> 0) & 0xff;
        int b2 = (value >> 8) & 0xff;

        return (ushort) (b1 << 8 | b2 << 0);
    });

    // Swaps 4 byte values
    vars.SwapBytesInt = (Func<uint, uint>)((value) => {
        return ((value & 0x000000ff) << 24) +
            ((value & 0x0000ff00) << 8) +
            ((value & 0x00ff0000) >> 8) +
            ((value & 0xff000000) >> 24);
    });

	settings.Add("events", true, "Events");

	vars.AddSplit("gmutant", true, "G-Mutant", "events");
	settings.SetToolTip("gmutant", "Splits when killed G-Mutant");

	vars.AddSplit("gleech", true, "G.Leech", "events");
	settings.SetToolTip("gleech", "Splits when killed G.Leech");

	vars.AddSplit("suspended", true, "Suspended", "events");
	settings.SetToolTip("suspended", "Splits when killed Suspended");

	vars.AddSplit("thanatos", true, "Thanatos", "events");
	settings.SetToolTip("thanatos", "Splits when killed Thanatos");

	vars.AddSplit("thanatosr", true, "Thanatos.R", "events");
	settings.SetToolTip("thanatosr", "Splits when killed Thanatos.R");

	vars.AddSplit("titan", true, "Titan", "events");
	settings.SetToolTip("titan", "Splits when killed Titan");

	vars.AddSplit("stalker", true, "Stalker", "events");
	settings.SetToolTip("stalker", "Splits when killed Stalker");

	vars.AddSplit("gigabyte", true, "Gigabyte", "events");
	settings.SetToolTip("gigabyte", "Splits when killed Gigabyte");

	vars.AddSplit("tyrantr", true, "Tyrant R", "events");
	settings.SetToolTip("tyrantr", "Splits when killed Tyrant R");

	vars.AddSplit("nyx", true, "Nyx", "events");
	settings.SetToolTip("nyx", "Splits when killed Nyx");

	vars.AddSplit("endGame", true, "End Game", "events");
	settings.SetToolTip("endGame", "Splits when the game rank screen displays");

}

init
{
    vars.lastSplit = String.Empty; // Name of the most recent split
    vars.gameProcess = String.Empty; // Used to detect when the process has changed
    vars.productCode = String.Empty; // Used to detect when the game release has changed
    vars.basePointer = IntPtr.Zero; // Emulator virtual memory base pointer
    vars.isBigEndian = false; // Console uses big endian (e.g. PS3 and GCN)

    // Log splits
    vars.LogSplit = (Action<string>)((text) => {
        vars.lastSplit = text;

        string log = timer.CurrentTime[timer.CurrentTimingMethod] + ": " + text;

        if (settings["logSplits"])
            File.AppendAllText(vars.logPath, log + System.Environment.NewLine);

        if (settings["debugSplits"])
            print("recvx split: " + log);
    });

    // Update split status
    vars.UpdateSplit = (Func<string, bool>)((split) => {
        if (!vars.splits[split])
        {
            vars.LogSplit(split);
            vars.splits[split] = true;
            return settings[split];
        }
        else
            return false;
    });

    // Update item split status
    vars.UpdateItemSplit = (Func<byte, bool>)((id) => {
        if (!vars.itemSplits[id])
        {
            vars.LogSplit(vars.itemIndex[id] + " (" + id.ToString() + ")");
            vars.itemSplits[id] = true;
            return settings[vars.itemIndex[id]];
        }
        else
            return false;
    });

    // Sets memory pointers for the detected game release
    vars.UpdatePointers = (Action) (() => {
        switch ((string)vars.productCode)
        {
            default: // SLPM_656.92 [PS2] [JP] Biohazard: outbreak f2
				vars.scenarioPtr = 0x003137BA;
                vars.timePtr = 0x004912B8;
                vars.roomPtr = 0x0047BF12;
                vars.screenPtr = 0x004912A0;
				vars.winPtr = 0x00491346;
                vars.healthPtr = 0x0047C270;
                vars.statusPtr = 0x0047C8E9;
                vars.pausePtr = 0x004912BE;
                vars.characterPtr = 0x0047C8E0;
                vars.inventoryPtr = 0x0047C9C4;
                vars.moveablePtr = 0x0047C918;
                vars.boss1hpPtr = 0x0042C640; //G-Mutant
                vars.boss2hpPtr = 0x0042E1A0; //G.Leech
                vars.boss3hpPtr = 0x0042E1A0; //Suspended
                vars.boss4hpPtr = 0x0042E1A0; //Thanatos
                vars.boss5hpPtr = 0x0042E1A0; //Thanatos.R
                vars.boss6hpPtr = 0x0042E1A0; //Titan
                vars.boss7hpPtr = 0x0042E1A0; //Stalker
                vars.boss8hpPtr = 0x0042E1A0; //Gigabyte
                vars.boss9hpPtr = 0x0042E1A0; //Tyrant R
                vars.boss10hpPtr = 0x0042E1A0; //Nyx
                vars.boss11hpPtr = 0x0042E1A0; //Tyrant slow
                vars.boss12hpPtr = 0x0042E1A0; //Tyrant fast
                vars.boss13hpPtr = 0x0042E1A0; //Tyrant red
                break;
        }
    });
	
    // Detects if the process has changed
    vars.UpdateProcess = (Action) (() => {
        // Update process name if application has changed
        if (vars.gameProcess != game.ProcessName)
        {
            vars.gameProcess = game.ProcessName;

            // Set emulator edianess
            switch ((string)vars.gameProcess.ToLower())
            {
                default: //pcsx2
                    vars.isBigEndian = false;
                    break;
            }
        }
    });

    // Set emulator base pointer
    vars.UpdatePointer = (Action) (() => {
        switch ((string)vars.gameProcess.ToLower())
        {
            default: // pcsx2
                vars.basePointer = 0x20000000;
                break;
        }
    });

    // Detects if the game release has changed
    vars.UpdateProduct = (Action) (() => {
        string productCode;

        // Read game product code from memory
        switch ((string)vars.gameProcess.ToLower())
        {
            default: //pcsx2
                productCode = memory.ReadString(new IntPtr(vars.basePointer + 0x00015B90), 11);
                break;
        }

        // Update product code if the game release has changed
        if (vars.productCode != productCode)
        {
            vars.productCode = productCode;
            vars.UpdatePointers();
        }
    });

    // Updates game values
    vars.UpdateValues = (Action) (() => {
        ushort scenario = 0; // Scenario ID
        uint time = 0; // In Game Timer
        byte room = 0; // Room the character is in
        byte screen = 0; // Current game screen
        ushort win = 0; // win result (0x0001 win, 0x0101 fail)
        uint health = 0; // Character health
        byte status = 0; // Character status info
        byte pause = 0; // pause status info
        byte character = 0; // Character ID
        byte moveable = 0; // When character can move
        ushort boss1hp = 0;
        ushort boss2hp = 0;
        ushort boss3hp = 0;
        ushort boss4hp = 0;
        ushort boss5hp = 0;
        ushort boss6hp = 0;
        ushort boss7hp = 0;
        ushort boss8hp = 0;
        ushort boss9hp = 0;
        ushort boss10hp = 0;
        ushort boss11hp = 0;
        ushort boss12hp = 0;
        ushort boss13hp = 0;

        // Read values from memory
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.scenarioPtr), out scenario);
        memory.ReadValue<uint>(new IntPtr(vars.basePointer + vars.timePtr), out time);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.roomPtr), out room);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.screenPtr), out screen);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.winPtr), out win);
        memory.ReadValue<uint>(new IntPtr(vars.basePointer + vars.healthPtr), out health);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.statusPtr), out status);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.pausePtr), out pause);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.characterPtr), out character);
        memory.ReadValue<byte>(new IntPtr(vars.basePointer + vars.moveablePtr), out moveable);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss1hpPtr), out boss1hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss2hpPtr), out boss2hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss3hpPtr), out boss3hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss4hpPtr), out boss4hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss5hpPtr), out boss5hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss6hpPtr), out boss6hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss7hpPtr), out boss7hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss8hpPtr), out boss8hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss9hpPtr), out boss9hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss10hpPtr), out boss10hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss11hpPtr), out boss11hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss12hpPtr), out boss12hp);
        memory.ReadValue<ushort>(new IntPtr(vars.basePointer + vars.boss13hpPtr), out boss13hp);

        current.slot = 0; // Inventory slot number of the equipped item
        current.ammo = 0; // Ammo count for the equipped weapon
        current.scenario = scenario;
		current.time = vars.isBigEndian ? (int)vars.SwapBytesInt(time) : (int)time;
        current.room = room;
        current.screen = screen;
        current.win = win;
        current.health = vars.isBigEndian ? (int)vars.SwapBytesInt(health) : (int)health;
        current.status = status;
        current.character = character;
        current.inventory = new byte[8]; // Current characters inventory
		current.moveable = moveable;
		current.boss1hp = boss1hp;
		current.boss2hp = boss2hp;
		current.boss3hp = boss3hp;
		current.boss4hp = boss4hp;
		current.boss5hp = boss5hp;
		current.boss6hp = boss6hp;
		current.boss7hp = boss7hp;
		current.boss8hp = boss8hp;
		current.boss9hp = boss9hp;
		current.boss10hp = boss10hp;
		current.boss11hp = boss11hp;
		current.boss12hp = boss12hp;
		current.boss13hp = boss13hp;

        int index = -1; // Inventory array index

        // Pointer to the current characters inventory table
        IntPtr pointer = new IntPtr(vars.basePointer + vars.inventoryPtr);

        // Read inventory of the current character
        for (int i = 0; i < 8; ++i)
        {
            byte item = 0; // Inventory entry

            // Read inventory entry from memory
            memory.ReadValue<byte>(pointer, out item);

            // Move pointer to the next entry
            pointer = IntPtr.Add(pointer, 0x1);

            // First entry is the slot number of the equipped item
            if (i == -1)
            {
                //current.slot = item;
                continue;
            }

            // Set inventory index to the item ID
            current.inventory[++index] = item;
        }
    });

    // Initialise values
    vars.UpdateProcess();
    vars.UpdatePointer();
    vars.UpdateProduct();
    vars.UpdateValues();
}

start
{
    // Start when IGT is 0, in the Prison Cell room and the 1st item slot contains the Lighter
    return (current.time > 0);
}

update
{
    vars.UpdateProcess();
    vars.UpdatePointer();
    vars.UpdateProduct();
    vars.UpdateValues();

    if (timer.CurrentPhase == TimerPhase.NotRunning)
    {
        vars.ResetSplits();
        vars.ResetItemSplits();
    }
}

split
{
    // Split when an inventory entry changes
    for (int i = 0; i < current.inventory.Length; i++)
    {
        byte item = current.inventory[i];

        // Continue if the entry has not changed
        if (item == old.inventory[i]) continue;

        switch (item)
        {
            default: // Split the first time an item is added to the inventory
                if (vars.itemSplits.ContainsKey(item))
                    return vars.UpdateItemSplit(item);
                break;
        }
    }

	if (current.room == 1 && (current.boss1hp >= 0x8000 || current.boss1hp == 0)) // when boss ded
		return vars.UpdateSplit("gmutant");

	if (current.room == 2 && (current.boss2hp >= 0x8000 || current.boss2hp == 0)) // when boss ded
		return vars.UpdateSplit("gleech");

	if (current.room == 3 && (current.boss3hp >= 0x8000 || current.boss3hp == 0)) // when boss ded
		return vars.UpdateSplit("suspended");

	if (current.room == 4 && (current.boss4hp >= 0x8000 || current.boss4hp == 0)) // when boss ded
		return vars.UpdateSplit("thanatos");

	if (current.room == 5 && (current.boss5hp >= 0x8000 || current.boss5hp == 0)) // when boss ded
		return vars.UpdateSplit("thanatosr");

	if (current.room == 6 && (current.boss6hp >= 0x8000 || current.boss6hp == 0)) // when boss ded
		return vars.UpdateSplit("titan");

	if (current.room == 7 && (current.boss7hp >= 0x8000 || current.boss7hp == 0)) // when boss ded
		return vars.UpdateSplit("stalker");

    if (current.room == 8 && (current.boss8hp >= 0x8000 || current.boss8hp == 0)) // when boss ded
        return vars.UpdateSplit("gigabyte");

	if (current.room == 9 && (current.boss9hp >= 0x8000 || current.boss9hp == 0)) // when boss ded
		return vars.UpdateSplit("tyrantr");

	if (current.room == 10 && (current.boss10hp >= 0x8000 || current.boss10hp == 0)) // when boss ded
		return vars.UpdateSplit("nyx");

    if (current.screen == 0x0C) // End Game Rank Screen
		return vars.UpdateSplit("endGame");
}

reset
{
    if (current.room == 0 && current.time == 0)
    {
        return true;
    }
}

gameTime
{
    // Seconds = IGT / 60
    return TimeSpan.FromSeconds(current.time / 0.03 / 1000);
}

isLoading
{
    return true;
}